from ewah.operators.sql_base_operator import EWAHSQLBaseOperator
from ewah.constants import EWAHConstants as EC

from mysql.connector import connect

from airflow.hooks.base_hook import BaseHook

class EWAHMySQLOperator(EWAHSQLBaseOperator):

    _SQL_BASE = \
        'SELECT\n{columns}\nFROM `{schema}`.`{table}`\nWHERE {where_clause}'
    _SQL_BASE_SELECT = \
        'SELECT * FROM ({select_sql}) t WHERE {{0}}'
    _SQL_COLUMN_QUOTE = '`'
    _SQL_MINMAX_CHUNKS = 'SELECT MIN({column}), MAX({column}) FROM ({base}) t;'
    _SQL_CHUNKING_CLAUSE = '''
        AND {column} >= %(from_value)s
        AND {column} <{equal_sign} %(until_value)s
    '''
    _SQL_PARAMS = '%({0})s'

    def __init__(self, *args, **kwargs):
        self.sql_engine = self._MYSQL
        super().__init__(*args, **kwargs)

    def _get_data_from_sql(self, sql, params=None, return_dict=True):
        connection = BaseHook.get_connection(self.source_conn_id)
        database_conn = connect(**{
            'host': connection.host,
            'user': connection.login,
            'passwd': connection.password,
            'port': connection.port,
            'database': connection.schema,
        })
        cursor = database_conn.cursor(dictionary=return_dict)
        self.log.info('Executing:\n{0}\n\nWith params:\n{1}'.format(
            sql,
            str(params),
        ))
        cursor.execute(sql, params=params)
        data = cursor.fetchall()
        cursor.close()
        database_conn.close()
        return data



from ewah.operators.base_operator import EWAHBaseOperator
from ewah.ewah_utils.airflow_utils import airflow_datetime_adjustments
from ewah.constants import EWAHConstants as EC

from airflow.hooks.base_hook import BaseHook

from facebook_business.api import FacebookAdsApi
from facebook_business.adobjects.adaccount import AdAccount
from facebook_business.adobjects.adsinsights import AdsInsights

from datetime import datetime, timedelta

import inspect
import time

class EWAHFBOperator(EWAHBaseOperator):

    template_fields = ('data_from', 'data_until')

    _IS_INCREMENTAL = True
    _IS_FULL_REFRESH = False

    class levels:
        ad = 'ad'

    def __init__(
        self,
        account_ids,
        insight_fields,
        level,
        data_from=None,
        data_until=None,
        time_increment=1,
        breakdowns=None,
        execution_waittime_seconds=15, # wait for a while before execution
        #   between account_ids to avoid hitting rate limits during backfill
        pagination_limit=1000,
        async_job_read_frequency_seconds=5,
        reload_data_from=None,
    *args, **kwargs):

        if kwargs.get('update_on_columns'):
            raise Exception('update_on_columns is set by operator!')

        if not account_ids.__iter__:
            raise Exception('account_ids must be an iterable, such as a list,' \
                + ' of strings or integers!')

        if level == self.levels.ad:
            kwargs['update_on_columns'] = [
                'ad_id',
                'date_start',
                'date_stop',
            ] + (breakdowns or [])
            insight_fields += ['ad_id', 'ad_name']
            insight_fields = list(set(insight_fields))
        else:
            raise Exception('Specified level not supported!')

        if not (
            (
                type(time_increment) == str
                and time_increment in ['monthly', 'all_days']
            )
            or
            (
                type(time_increment) == int
                and time_increment >= 1
                and time_increment <= 1
            )
        ):
            raise Exception('time_increment must either be an integer ' \
                + 'between 1 and 90, or a string of either "monthly" '\
                + 'or "all_days". Recommended and default is the integer 1.')

        allowed_insight_fields = [
            _attr[1] for _attr in [
                member for member in inspect.getmembers(
                    AdsInsights.Field,
                    lambda a:not (inspect.isroutine(a)),
                )
                if not (member[0].startswith('__') and member[0].endswith('__'))
            ]
        ]
        for i_f in insight_fields:
            if not i_f in allowed_insight_fields:
                raise Exception((
                    'Field {0} is not an accepted value for insight_fields! ' \
                    + 'Accepted field values:\n\t{1}\n'
                ).format(
                    i_f,
                    '\n\t'.join(allowed_insight_fields)
                ))

        super().__init__(*args, **kwargs)

        credentials = BaseHook.get_connection(self.source_conn_id)
        extra = credentials.extra_dejson

        # Note: app_secret is not always required!
        if not extra.get('app_id'):
            raise Exception('Connection extra must contain an "app_id"!')
        if not extra.get('access_token', credentials.password):
            raise Exception('Connection extra must contain an "access_token" ' \
                + 'if it is not saved as the connection password!')

        self.credentials = {
            'app_id': extra.get('app_id'),
            'app_secret': extra.get('app_secret'),
            'access_token': extra.get('access_token', credentials.password),
        }

        self.account_ids = account_ids
        self.insight_fields = insight_fields
        self.data_from = data_from
        self.data_until = data_until
        self.level = level
        self.time_increment = time_increment
        self.breakdowns = breakdowns
        self.execution_waittime_seconds = execution_waittime_seconds
        self.pagination_limit = pagination_limit
        self.async_job_read_frequency_seconds = async_job_read_frequency_seconds
        self.reload_data_from = reload_data_from

    def _clean_response_data(self, response):
        return [dict(datum) for datum in list(response)]

    def execute(self, context):
        if not self.test_if_target_table_exists():
            if self.reload_data_from:
                self.data_from = self.reload_data_from

        self.data_from = airflow_datetime_adjustments(self.data_from)
        self.data_until = airflow_datetime_adjustments(self.data_until)

        self.data_from = self.data_from or context['execution_date']
        self.data_until = self.data_until or context['next_execution_date']

        time_range = {
            'since': self.data_from.strftime('%Y-%m-%d'),
            'until': self.data_until.strftime('%Y-%m-%d'),
        }

        FacebookAdsApi.init(**self.credentials)
        params = {
            'time_range': time_range,
            'time_increment': self.time_increment,
            'level': self.level,
            'limit': self.pagination_limit,
        }
        if self.breakdowns:
            params.update({'breakdowns': ','.join(self.breakdowns)})

        for account_id in self.account_ids:
            if self.execution_waittime_seconds:
                self.log.info('Delaying execution by {0} seconds...'.format(
                    str(self.execution_waittime_seconds),
                ))
                now = datetime.now()
                while datetime.now() < \
                    (now + timedelta(seconds=self.execution_waittime_seconds)):
                    time.sleep(1)

            account_object = AdAccount('act_{0}'.format(str(account_id)))
            self.log.info((
                'Requesting data for account_id={0} between {1} and {2}.'
            ).format(
                str(account_id),
                time_range['since'],
                time_range['until'],
            ))

            async_job = account_object.get_insights_async(
                fields=self.insight_fields,
                params=params,
            )
            job_remote_read = async_job.api_get()
            done_status = [
                'Job Completed',
                'Job Failed',
                'Job Skipped',
            ]
            while not (job_remote_read.get('async_status') in done_status):
                self.log.info('Asnyc job completion: {0}% (status: {1})'.format(
                    str(job_remote_read.get('async_percent_completion')),
                    str(job_remote_read.get('async_status')),
                ))
                time.sleep(self.async_job_read_frequency_seconds)
                job_remote_read = async_job.api_get()

            time.sleep(1)
            assert job_remote_read.get('async_status') == 'Job Completed'
            data = self._clean_response_data(async_job.get_result(
                params={'limit': self.pagination_limit},
            ))
            self.upload_data(data)



from airflow.models import BaseOperator
from airflow.hooks.base_hook import BaseHook

from ewah.dwhooks import get_dwhook
from ewah.constants import EWAHConstants as EC


class EWAHBaseOperator(BaseOperator):
    """Extension of airflow's native Base Operator.

    EWAH operators always work in the same way: They extract raw data from a
    source and load it into a relational database aka data warehouse aka DWH.
    Transformations only happen insofar as they are required to bring data
    into a relational format.

    *How to use*
    The child class extracts data from the source. At the end of the child
    class' execute() function, it calls self.update(). Optional: The child can
    also repeatedly call the self.update() class for chunking.

    *Arguments of self.update():*
    Data is a list of dictionaries, where each list element will become
    one row of data and the contained dict is fieldname:value. Value can be
    None, or a fieldname can be missing in a particular dictionary, in which
    case the value is assumed to be None.

    columns_definition may be supplied at the self.update() function call,
    at initialization, or not at all, depending on the usecase. The use order is
    - If available, take the columns_definition from the self.update() call
    - If None, check if columns_definition was supplied to the operator at init
    - If neither, create a columns_definition on the fly using (all) the data

    columns_definition is a dictionary with fieldname:properties. Properties
    is None or a dictionary of option:value where option can be one of the
    following:
    - EC.QBC_FIELD_TYPE -> String: Field type (default text)
    - EC.QBC_FIELD_PK -> Boolean: Is this field the primary key? (default False)
    - EC.QBC_FIELD_NN -> Boolean: Not Null constraint for this field? (False)
    - EC.QBC_FIELD_UQ -> Boolean: Unique constraint for this field? (False)

    Note that the value of EC.QBC_FIELD_TYPE is DWH-engine specific!

    columns_definition is case sensitive.

    Implemented DWH engines:
    - PostgreSQL
    - Snowflake

    Next up:
    - BigQuery
    - Redshift
    """

    _IS_INCREMENTAL = False # Child class must update these values accordingly.
    _IS_FULL_REFRESH = False # Defines whether operator is usable in factories.

    _REQUIRES_COLUMNS_DEFINITION = False # raise error if true an none supplied

    upload_call_count = 0

    _metadata = {} # to be updated by operator, if applicable

    def __init__(
        self,
        source_conn_id,
        dwh_engine,
        dwh_conn_id,
        target_table_name,
        target_schema_name,
        target_schema_suffix='_next',
        target_database_name=None, # Only for Snowflake
        columns_definition=None,
        drop_and_replace=True,
        update_on_columns=None,
        primary_key_column_name=None,
        clean_data_before_upload=True,
        add_metadata=True, # adds columns with metadata to all rows
        # Note: that metadata is specified by a dict on operator level!
        # metadata can only be added if no columns definition is given
    *args, **kwargs):

        if not dwh_engine or not dwh_engine in EC.DWH_ENGINES:
            raise Exception('Invalid DWH Engine: {0}\n\nAccapted Engines:{1}'
                .format(
                str(dwh_engine),
                '\n'.join(EC.DWH_ENGINES),
            ))

        if dwh_engine == EC.DWH_ENGINE_SNOWFLAKE:
            if not target_database_name:
                conn_db_name = BaseHook.get_connection(dwh_conn_id)
                conn_db_name = conn_db_name.extra_dejson.get('database')
                if conn_db_name:
                    target_database_name = conn_db_name
                else:
                    raise Exception('If using DWH Engine {0}, must provide {1}!'
                        .format(
                            dwh_engine,
                            '"target_database_name" to specify the Database',
                        )
                    )
        else:
            if target_database_name:
                raise Exception('Received argument for "target_database_name"!')

        if self._REQUIRES_COLUMNS_DEFINITION:
            if not columns_definition:
                raise Exception('This operator requires the argument ' \
                    + 'columns_definition!')

        if not drop_and_replace:
            if not (
                update_on_columns
                or primary_key_column_name
                or (columns_definition and (0 < sum([
                        bool(columns_definition[col].get(
                            EC.QBC_FIELD_PK
                        )) for col in list(columns_definition.keys())
                    ])))
                ):
                raise Exception("If this is incremental loading of a table, "
                    + "one of the following is required:"
                    + "\n- List of columns to update on (update_on_columns)"
                    + "\n- Name of the primary key (primary_key_column_name)"
                    + "\n- Column definition (columns_definition) that includes"
                    + " the primary key(s)"
                )

        super().__init__(*args, **kwargs)

        self.source_conn_id = source_conn_id
        self.dwh_engine = dwh_engine
        self.dwh_conn_id = dwh_conn_id
        self.target_table_name = target_table_name
        self.target_schema_name = target_schema_name
        self.target_schema_suffix = target_schema_suffix
        self.target_database_name = target_database_name
        self.columns_definition = columns_definition
        self.drop_and_replace = drop_and_replace
        if (not update_on_columns) and primary_key_column_name:
            if type(primary_key_column_name) == str:
                update_on_columns = [primary_key_column_name]
            elif type(primary_key_column_name) in (list, tuple):
                update_on_columns = primary_key_column_name
        self.update_on_columns = update_on_columns
        self.clean_data_before_upload = clean_data_before_upload
        self.primary_key_column_name = primary_key_column_name # may be used ...
        #   ... by a child class at execution!
        self.add_metadata = add_metadata

        self.hook = get_dwhook(self.dwh_engine)

        # wrap stuff around the final execute function, including the commit
        self.execute = self.wrap_exec(self.execute)

    def wrap_exec(self, exec_func):
        def callable_func(self=self, *args, **kwargs):
            self.upload_hook = self.hook(self.dwh_conn_id)
            # execute operator
            result = exec_func(*args, **kwargs)
            self.log.info('Now committing changes!')
            # commit only at the end, when all has worked out!
            self.upload_hook.commit()
            self.upload_hook.close()
            return result
        return callable_func

    def test_if_target_table_exists(self):
        hook = self.hook(self.dwh_conn_id)
        if self.dwh_engine == EC.DWH_ENGINE_POSTGRES:
            result = hook.test_if_table_exists(
                table_name=self.target_table_name,
                schema_name=self.target_schema_name + self.target_schema_suffix,
            )
            hook.close()
            return result
        elif self.dwh_engine == EC.DWH_ENGINE_SNOWFLAKE:
            result = hook.test_if_table_exists(
                table_name=self.target_table_name,
                schema_name=self.target_schema_name + self.target_schema_suffix,
                database_name=self.target_database_name,
            )
            hook.close()
            return result
        raise Exception('Function not implemented for DWH {0}!'.format(
            dwh_engine
        ))

    def _create_columns_definition(self, data):
        "Create a columns_definition from data (list of dicts)."
        inconsistent_data_type = EC.QBC_TYPE_MAPPING[self.dwh_engine].get(
            EC.QBC_TYPE_MAPPING_INCONSISTENT
        )
        def get_field_type(value):
            return EC.QBC_TYPE_MAPPING[self.dwh_engine].get(
                type(value)
            ) or inconsistent_data_type

        result = {}
        for datum in data:
            if self.add_metadata and self._metadata:
                datum.update(self._metadata)
            for field in datum.keys():
                if not (result.get(field, {}).get(EC.QBC_FIELD_TYPE) \
                    == inconsistent_data_type) and (not datum[field] is None):
                    if result.get(field):
                        # column has been added in a previous iteration.
                        # If not default column: check if new and old column
                        #   type identification agree.
                        if not (result[field][EC.QBC_FIELD_TYPE] \
                            == get_field_type(datum[field])):
                            self.log.info(
                                'WARNING! Data types are inconsistent.'
                                + ' Affected column: {0}'.format(field)
                            )
                            result[field][EC.QBC_FIELD_TYPE] = \
                                inconsistent_data_type

                    else:
                        # First iteration with this column. Add to result.
                        result.update({field:{
                            EC.QBC_FIELD_TYPE: get_field_type(datum[field])
                        }})
        return result

    def upload_data(
        self,
        data=None,
        columns_definition=None,
    ):
        if not data:
            self.log.info('No data to upload!')
            return
        self.upload_call_count += 1
        self.log.info('Chunk {1}: Uploading {0} rows of data.'.format(
            str(len(data)),
            str(self.upload_call_count),
        ))

        columns_definition = columns_definition or self.columns_definition
        if not columns_definition:
            self.log.info('Creating table schema on the fly based on data.')
            # Note: This is also where metadata is added, if applicable
            columns_definition = self._create_columns_definition(data)

        hook = self.upload_hook

        if (not self.drop_and_replace) or (self.upload_call_count > 1):
            self.log.info('Checking for, and applying schema changes.')
            new_cols, del_cols = hook.detect_and_apply_schema_changes(
                new_schema_name=self.target_schema_name+self.target_schema_suffix,
                new_table_name=self.target_table_name,
                new_columns_dictionary=columns_definition,
                # When introducing a feature utilizing this, remember to
                #  consider multiple runs within the same execution
                drop_missing_columns=False and self.upload_call_count==1,
                database=self.target_database_name,
                commit=False, # Commit only when / after uploading data
            )
            self.log.info('Added fields:\n\t{0}\nDeleted fields:\n\t{1}'.format(
                '\n\t'.join(new_cols) or '\n',
                '\n\t'.join(del_cols) or '\n',
            ))

        self.log.info('Uploading data now.')
        hook.create_or_update_table(
            data=data,
            columns_definition=columns_definition,
            table_name=self.target_table_name,
            schema_name=self.target_schema_name+self.target_schema_suffix,
            database_name=self.target_database_name,
            drop_and_replace=self.drop_and_replace and \
                (self.upload_call_count == 1), # In case of chunking of uploads
            update_on_columns=self.update_on_columns,
            commit=False,
            logging_function=self.log.info,
            clean_data_before_upload=self.clean_data_before_upload,
        )

        #hook.commit()
        #hook.close() conn is committed closed at the end by the wrapper func!

class EWAHEmptyOperator(EWAHBaseOperator):
    _IS_INCREMENTAL = True
    _IS_FULL_REFRESH = True
    def __init__(self, *args, **kwargs):
        raise Exception('Failed to load operator! Probably missing' \
            + ' requirements for the operator in question.\n\nSupplied args:' \
            + '\n\t' + '\n\t'.join(args) + '\n\nSupplied kwargs:\n\t' \
            + '\n\t'.join(['{0}: {1}'.format(k, v) for k, v in kwargs.items()])
        )



from ewah.operators.base_operator import EWAHBaseOperator
from ewah.ewah_utils.airflow_utils import airflow_datetime_adjustments
from ewah.constants import EWAHConstants as EC

from airflow.hooks.base_hook import BaseHook

from datetime import datetime, timedelta

from requests.auth import HTTPBasicAuth
import requests
import json
import time

class EWAHZendeskOperator(EWAHBaseOperator):

    _IS_INCREMENTAL = True
    _IS_FULL_REFRESH = False

    _base_url = \
        'https://{support_url}.zendesk.com/api/v2/incremental/{resource}.json'

    _accepted_resources = [
        'tickets',
        'ticket_metric_events',
        'users',
    ]
    def __init__(self,
        support_url,
        resource,
        auth_type,
        data_from,
    *args, **kwargs):

        if not resource in self._accepted_resources:
            raise Exception('resource "{0}" not supported!'.format(resource))

        if not auth_type in ['basic_auth']:
            raise Exception('auth_type must be basic_auth!')

        #if not type(page_limit) == int or page_limit < 1 or page_limit > 100:
        #    raise Exception('page_limit must be a positive interger <= 100!')

        if not type(data_from) == datetime:
            raise Exception('data_from must be a datetime object!')

        kwargs['primary_key_column_name'] = \
            kwargs.get('primary_key_column_name', 'id')

        super().__init__(*args, **kwargs)

        self.support_url = support_url
        self.resource = resource
        self.auth_type = auth_type
        self.data_from = data_from
        #self.page_limit = page_limit

    def execute(self, context):
        # Make sure it is at least 70 seconds after next_execution_date!
        td70 = timedelta(seconds=70)
        while datetime.now() < (context['next_execution_date'] + td70):
            self.log.info('delaying execution...')
            time.sleep(5)

        url = self._base_url.format(**{
            'support_url': self.support_url,
            'resource': self.resource,
        })

        if self.test_if_target_table_exists():
            data_from = context['execution_date']
        else:
            data_from = self.data_from
        data_from = data_from.timetuple()
        data_until = time.mktime(context['next_execution_date'].timetuple())

        params = {
            'start_time': str(time.mktime(data_from)),
        }

        conn = BaseHook.get_connection(self.source_conn_id)
        if self.auth_type == 'basic_auth':
            auth = HTTPBasicAuth(conn.login, conn.password)
        else:
            raise Exception('auth_type not implemented!')

        # Time-based incremental exports, run the first, the paginate!
        self.log.info('Calling Zendesk api.\nurl: {0}\nparams:{1}'.format(
            url,
            str(params),
        ))
        r = requests.get(url, params=params, auth=auth)
        data = json.loads(r.text)
        while r.status_code == 200 \
            and not data.get('end_of_stream') \
            and data.get('end_time') \
            and data.get('end_time') <= data_until:
            self.upload_data(data[self.resource]) # uploads previous request
            self.log.info('Requesting next page of data...')
            r = requests.get(data['next_page'], auth=auth) # new request
            data = json.loads(r.text)

        if not r.status_code == 200:
            raise Exception('Error {1} when calling Zendesk API: {0}'.format(
                r.text,
                str(r.status_code),
            ))

        # upload final request
        self.upload_data(data[self.resource])



from ewah.operators.base_operator import EWAHBaseOperator
from ewah.ewah_utils.airflow_utils import airflow_datetime_adjustments
from ewah.constants import EWAHConstants as EC

from airflow.hooks.S3_hook import S3Hook

from datetime import datetime, timedelta

import json

class EWAHS3Operator(EWAHBaseOperator):
    """Only implemented for JSON files from S3 right now!"""

    template_fields = ('data_from', 'data_until')

    _IS_INCREMENTAL = True
    _IS_FULL_REFRESH = True

    _IMPLEMENTED_FORMATS = ['JSON', 'AWS_FIREHOSE_JSON']

    def __init__(self,
        bucket_name,
        file_format,
        key_name=None, # either specify a key or specify data_from / data_until
        data_from=None,
        data_until=None,
    *args, **kwargs):

        if not file_format in self._IMPLEMENTED_FORMATS:
            raise Exception('File format not implemented! Available formats:' \
                + ' {0}'.format(str(self._IMPLEMENTED_FORMATS)))

        if key_name:
            if data_from or data_until:
                raise Exception('Both key_name and either data_from or data_' \
                    + 'until specified! Cannot have both specified.')

        super().__init__(*args, **kwargs)

        if not self.drop_and_replace:
            if not (data_from and data_until):
                raise Exception('For incremental loading, you must specify ' \
                    + 'both data_from and data_until')

        self.bucket_name = bucket_name
        self.key_name = key_name
        self.data_from = data_from
        self.data_until = data_until
        self.file_format = file_format

    def execute(self, context):
        if self.file_format == 'JSON':
            return self.execute_json(
                context=context,
                f_get_data=lambda hook, key, bucket: json.loads(
                    hook.read_key(key, bucket)
                ),
            )
        elif self.file_format == 'AWS_FIREHOSE_JSON':
            return self.execute_json(
                context=context,
                f_get_data=lambda hook, key, bucket: json.loads(
                    '['+hook.read_key(key, bucket).replace('}{', '},{')+']'
                ),
            )
        else:
            raise Exception('File format not implemented!')

    def execute_json(self, context, f_get_data):
        self.data_from = airflow_datetime_adjustments(self.data_from)
        self.data_until = airflow_datetime_adjustments(self.data_until)

        hook = S3Hook(self.source_conn_id)

        if self.key_name:
            data = f_get_data(
                hook=hook,
                key=self.key_name,
                bucket=self.bucket_name,
            )
        else:
            data = []
            bucket = hook.get_bucket(self.bucket_name)
            for obj in bucket.objects.all():
                if self.data_from and obj.last_modified < self.data_from:
                    continue
                if self.data_until and obj.last_modified >= self.data_until:
                    continue

                self.log.info('Loading data from file {0}'.format(
                    obj.key,
                ))
                data += f_get_data(
                    hook=hook,
                    key=obj.key,
                    bucket=self.bucket_name,
                )

        print(self.columns_definition)
        self.upload_data(data=data)



from ewah.constants import EWAHConstants as EC
from ewah.operators import *
from ewah.operators.base_operator import EWAHBaseOperator as EBO

from .dag_factory_full_refresh import dag_factory_drop_and_replace
from .dag_factory_incremental import dag_factory_incremental_loading

from copy import deepcopy

import yaml
from ewah.ewah_utils.yml_loader import Loader, Dumper

def dags_from_dict(
    dag_definition,
    raise_error_when_empty=False,
    raise_warnings_as_errors=False,
):
    """This function returns a list of airflow DAGs as specified by a dictionary
    containing all configuration of the DAGs.

    :param dag_definition:  Dictionary containing all configuration of the DAGs.
    :type dag_definition:   Dict
    :param raise_error_when_empty:  If no DAGs are configured, raise an error?
    :type raise_error_when_empty:   Boolean
    :param raise_warnings_as_errors:    If true, warnings become errors.
    :type raise_warnings_as_errors:     Boolean

    Levels of the dictionary:
    1) The first level may be one of the following:
    - el_dags: contains the actual DAG definitions
    - base_config: template configuration for all DAGs - will be overwritten
        by individual configurations, except for additional_task_args, which
        will only be updated by DAG configurations

    el_dags is expected to be a dictionary where the key is the (base) name of
        the DAG(s) and the value is a dictionary of its configuration.
    Configuration options:
        - incremental   is this incremental loading? Defaults to False
        - dwh_engine    What engine does it run on? Possible values:
            postgres, snowflake
        - dwh_conn_id   airflow connection name (id) for the target DWH
        - el_operator   name of the EWAH EL Operator, one of: fx, ga, mysql,
            oracle, postgres, s3
        - start_date    DAG start date
        - end_date      DAG end date
        - read_right_users  comma-separated string or list of users or roles
            to receive read rights on the created schemas and tables.
        - target_schema_name    name of the schema to load data into (attn:
            only one schema per DAG! Is not automatically checked!)
        - target_schema_suffix  optional, temporary suffix for replacement
            schemas while loading new data
        - target_database_name optional for snowflake DWHs, forbidden otherwise
            -> name of the target database (looks at dwh_conn_id's extra json
            if not specified for a snowflake DWH)
        - additional_task_args optional dict with arguments for the individual
            tasks of the DAG(s). Allowed arguments are:
            - email
            - email_on_retry
            - email_on_failure
            - retries
            - retry_delay
            - priorirty_weight
            - weight_rule
            - pool
            - owner
            ->> See airflow documentation for details
        - additional_dag_args optional dict, you can additionally set the
            following DAG-level arguments in this dict:
            - description
            - params
            - default_view
            - orientation
            - access_control
            ->> See airflow documentation for details
        - operator_config this is complex enough to warrant its own section

        for drop and replace only:
        - schedule_interval only for drop and replace: schedule interval of DAG
            - defaults to timedelta(days=1), give as a dict containing
            interval(s), e.g. days=1 or hours=3

        for incremental loading only:
        - airflow_conn_id airflow connection name (id) for the airflow metadata
            database
        - schedule_interval_backfill the schedule interval for the backfill DAG
            timedelta - use this notation:
            schedule_interval: !!python/object/apply:datetime.timedelta
                - days go first
                - seconds go second
                - microseconds go third
        - schedule_interval_future the schedule interval for the normal DAG
            also a timedelta
        - switch_date the date to switch from backfill to normal DAG - if
            ommitted, it is calculated automatically based on start_date
            and the schedule intervals to be the latest possible date
            that is still in the past

        The operator_config: This is where arguments are specified that are
            related to the specific operator in use. There are four parts,
            two of which are always relevant and two are unique to incremental
            DAGs: general_config, incremental_config, backfill_config and tables
        general_config
            Is at the lowest level of the four and all arguments here can be
            overwritten at one of the other levels. Usually, this is where one
            specifies a source_conn_id or data_from and data_until
        incremental_config
            This is config to apply only to normal DAGs but not backfill DAGs
            for incremental DAG sets. Overwrites general_config, is overwritten
            by table config
        backfill_config
            Same as incremental_config, but only applies to the backfill DAGs
        tables
            Is a dict of dicts, where the key of the dict is the name of the
            table in the DWH. May also contain any argument to overwrite the
            other operator config values as well as the arguments
            "skip_backfill", which removes this particular table from backfill
            DAGs, and "drop_and_replace", which is False per default but may
            be useful as True for some tables, especially used in conjunction
            with "skip_backfill"
    """

    _INCR_ONLY = [
        'airflow_conn_id',
        'schedule_interval_backfill',
        'schedule_interval_future',
        'switch_date',
    ]
    _FR_ONLY = ['schedule_interval']

    dwh_engines = {
        'postgres': EC.DWH_ENGINE_POSTGRES,
        'snowflake': EC.DWH_ENGINE_SNOWFLAKE,
        # 'bigquery': EC.DWH_ENGINE_BIGQUERY,
        # 'redshift': EC.DWH_ENGINE_REDSHIFT,
        # 's3': EC.DWH_ENGINE_S3,
    }

    operators = {
        'fx': EWAHFXOperator,
        'ga': EWAHGAOperator,
        'google_analytics': EWAHGAOperator,
        'mysql': EWAHMySQLOperator,
        'oracle': EWAHOracleSQLOperator,
        'postgres': EWAHPostgresOperator,
        'postgresql': EWAHPostgresOperator,
        'pgsql': EWAHPostgresOperator,
        's3': EWAHS3Operator,
        'fb': EWAHFBOperator,
        'facebook': EWAHFBOperator,
        'gsheets': EWAHGSpreadOperator,
        'google_sheets': EWAHGSpreadOperator,
        'gs': EWAHGSpreadOperator,
        'mongo': EWAHMongoDBOperator,
        'mongodb': EWAHMongoDBOperator,
        'shopify': EWAHShopifyOperator,
        'zendesk': EWAHZendeskOperator,
    }

    allowed_dag_args = [
        'description',
        'params',
        'default_view',
        'orientation',
        'access_control',
        'concurrency'
    ]

    allowed_task_args = [
        'email',
        'email_on_retry',
        'email_on_failure',
        'retries',
        'retry_delay',
        'priorirty_weight',
        'weight_rule',
        'pool',
        'owner', # set within this function, can be overwritten
    ]

    def warn_me(*warning_strings):
        if raise_warnings_as_errors:
            raise Exception(' '.join(warning_strings))
        print(*warning_strings)

    def check_keys_are_allowed(dict, allowed_list, dict_name):
        for key in dict.keys():
            if not key in allowed_list:
                warn_me('Unexpected key {0} was found in dict {1}!'.format(
                    key,
                    dict_name,
                ))

    base_config = dag_definition.pop('base_config', {})
    dags_dict = dag_definition.pop('el_dags', {})
    if not dags_dict:
        warn_me('No EL DAGs specified!')

    additional_dag_args = base_config.pop('additional_dag_args', {})
    base_config['additional_dag_args'] = {}
    for key, value in additional_dag_args.items():
        if key in allowed_dag_args:
            base_config['additional_dag_args'].update({key: value})
        else:
            warn_me(('Unexpected key {0} found in ' \
                + 'additional_dag_args in base_config!').format(
                key,
            ))

    additional_task_args = base_config.pop('additional_task_args', {})
    base_config['additional_task_args'] = {'owner': 'EWAH'}
    for key, value in additional_task_args.items():
        if key in allowed_task_args:
            base_config['additional_task_args'].update({key: value})
        else:
            warn_me('Unexpected key {0} found in additional_task_args!'.format(
                key,
            ))

    if dag_definition:
        warn_me(
            'Received unexpected arguments:',
            '\n{0}'.format(str(dag_definition)),
        )

    dags = []
    for dag_name, dag_config in dags_dict.items():
        config = deepcopy(base_config)
        config['additional_task_args'].update(
            dag_config.pop('additional_task_args', {})
        )
        config['additional_dag_args'].update(
            dag_config.pop('additional_dag_args', {})
        )
        check_keys_are_allowed(
            config['additional_task_args'],
            allowed_task_args,
            'el_dags.{0}.additional_task_args'.format(dag_name),
        )
        check_keys_are_allowed(
            config['additional_dag_args'],
            allowed_dag_args,
            'el_dags.{0}.additional_dag_args'.format(dag_name),
        )
        config.update(dag_config)
        drop_and_replace = not config.pop('incremental', False)
        try:
            config.update({'dwh_engine': dwh_engines[config['dwh_engine']]})
        except KeyError:
            raise Exception('Invalid dwh_engine {0}\nAllowed:\n\t{1}'.format(
                str(config.get('dwh_engine', None)),
                '\n\t'.join(list(dwh_engines.keys())),
            ))
        try:
            if type(config['el_operator']) == type \
                and issubclass(config['el_operator'], EBO):
                # custom operator was defined in yml file, use it
                config.update({
                    'el_operator': config['el_operator'],
                })
            else:
                config.update({
                    'el_operator': operators[config['el_operator'].lower()],
                })
        except KeyError:
            raise Exception('Invalid el_operator {0}\nAllowed:\n\t{1}'.format(
                str(config.get('el_operator', None)),
                '\n\t'.join(list(operators.keys())),
            ))

        if drop_and_replace:
            config.update({'dag_name': dag_name})
            for key in _INCR_ONLY: # allow specifying everything in base config
                config.pop(key, '')
            dags += [dag_factory_drop_and_replace(**config)]
        else:
            for key in _FR_ONLY:
                config.pop(key, '')
            config.update({'dag_base_name': dag_name})
            dags += list(dag_factory_incremental_loading(**config))
    if dags:
        return dags
    else:
        warn_me('No DAGs created!')
        return []

def dags_from_yml_file(
    file_path,
    raise_error_when_empty=True,
    raise_warnings_as_errors=True,
):

    return dags_from_dict(
        dag_definition=yaml.load(open(file_path, 'r'), Loader=Loader),
        raise_error_when_empty=raise_error_when_empty,
        raise_warnings_as_errors=raise_warnings_as_errors,
    )



from ewah.operators.base_operator import EWAHBaseOperator
from ewah.ewah_utils.airflow_utils import airflow_datetime_adjustments
from ewah.constants import EWAHConstants as EC

from airflow.hooks.base_hook import BaseHook

import json
from datetime import timedelta
from pymongo import MongoClient
from pymongo import ASCENDING as asc
from bson.json_util import dumps
from copy import deepcopy

class EWAHMongoDBOperator(EWAHBaseOperator):

    template_fields = ('data_from', 'data_until', 'reload_data_from')

    _IS_INCREMENTAL = True
    _IS_FULL_REFRESH = True

    _REQUIRES_COLUMNS_DEFINITION = False

    def __init__(self,
        source_collection_name=None, # defaults to target_table_name
        source_database_name=None, # Not required if specified in connection uri
        data_from=None, # string (can be templated) or datetime
        data_until=None, # as data_from
        timestamp_field=None, # required for use with data_from and data_until
        chunking_field=None, # defaults to timestamp_field if None
        chunking_interval=None, # timedelta or integer
        reload_data_from=None, # string (can be templated), datetime or None
        pagination_limit=None, # integer
        single_column_mode=False, # If True, throw all data as json into one col
    *args, **kwargs):

        src = source_collection_name or kwargs.get('target_table_name')
        self.source_collection_name = src
        self.source_database_name = source_database_name

        if data_from or data_until or reload_data_from:
            if not timestamp_field:
                err_msg = 'If using reload_data_from, data_from, or data_until, '
                err_msg += 'you must also specify timestamp_field!'
                raise Exception(err_msg)
        self.timestamp_field = timestamp_field
        self.data_from = data_from
        self.data_until = data_until
        self.reload_data_from = reload_data_from

        chunking_field = chunking_field or timestamp_field
        if chunking_interval:
            if not (type(chunking_interval) in [int, timedelta]):
                raise Exception('chunking_interval must be int or timedelta!')
            if not chunking_field:
                raise Exception('if chunking_interval is set, must also set ' \
                    'chunking_field or timestamp_field!')
        self.chunking_interval = chunking_interval
        self.chunking_field = chunking_field

        if pagination_limit:
            if not type(pagination_limit) == int:
                raise Exception('pagination_limit must be an integer!')
            if not kwargs.get('primary_key_column_name'):
                raise Exception('pagination_limit must be used with ' \
                    + 'primary_key_column_name!')

        self.pagination_limit = pagination_limit

        if single_column_mode:
            if kwargs.get('columns_definition'):
                raise Exception('single_column_mode is not compatible with '\
                    + 'columns_definition!')
            if not kwargs.get('drop_and_replace', True):
                raise Exception('single_column_mode is only compatible with ' \
                    + 'drop_and_replace!')
            if kwargs.get('update_on_columns'):
                raise Exception('single_column_mode is not compatible with ' \
                    + 'update_on_columns!')
            if kwargs.get('primary_key_column_name'):
                raise Exception('single_column_mode is not compatible with ' \
                    + 'primary_key_column_name!')
        self.single_column_mode = single_column_mode

        super().__init__(*args, **kwargs)

    def upload_data(self, data):
        if self.single_column_mode:
            # load all data into a single column called 'document'
            data = [{'document':x} for x in data]
            super().upload_data(data)
        else:
            # normal data loading mode
            super().upload_data(data)

    def extract_and_load_paginated(self,
        collection,
        filter_expressions=None,
        page_size=None,
        last_id=None,
    ):
        # adapted from https://www.codementor.io/@arpitbhayani/fast-and-efficient-pagination-in-mongodb-9095flbqr
        if page_size is None:
            # just get all data at once without pagination
            self.upload_data(
                json.loads(dumps(collection.find(filter_expressions)))
            )
            return

        # pagination!
        fe = deepcopy(filter_expressions)
        if last_id:
            # get next page
            fe.update({self.primary_key_column_name:{'$gt':last_id}})
        next_id_data = [x for x in collection.find(fe) \
                        .sort(self.primary_key_column_name, asc) \
                        .skip(page_size - 1)
                        .limit(1)
        ]
        data = json.loads(dumps(collection.find(fe)
                                    .sort(self.primary_key_column_name, asc) \
                                    .limit(page_size)
        ))

        if not len(data) == 0:
            self.upload_data(data)
            if not len(next_id_data) == 0:
                last_id = next_id_data[0][self.primary_key_column_name]
                # call recursively!
                self.extract_and_load_paginated(
                    collection=collection,
                    filter_expressions=filter_expressions,
                    page_size=page_size,
                    last_id=last_id,
                )

    def execute(self, context):
        if not self.drop_and_replace and not self.test_if_target_table_exists():
            self.data_from = self.reload_data_from
            self.log.info('Reloading data from {0}'.format(str(self.data_from)))
        self.data_from = airflow_datetime_adjustments(self.data_from)
        self.data_until = airflow_datetime_adjustments(self.data_until)

        self.log.info('Creating filter expression...')
        base_filters = []

        # data_from and data_until filter expression
        if self.timestamp_field:
            if self.data_from:
                base_filters += [{
                    self.timestamp_field: {'$gte': self.data_from}
                }]
            if self.data_until:
                base_filters += [{
                    self.timestamp_field: {'$lt': self.data_until}
                }]

        self.log.info('Connecting to MongoDB Database...')
        uri = BaseHook.get_connection(self.source_conn_id).password
        database = MongoClient(uri).get_database(name=self.source_database_name)

        self.log.info('Getting data...')
        collection = database[self.source_collection_name]
        if self.chunking_interval:
            # get data in chunks
            if base_filters:
                match_filters = {}
                for filter in base_filters:
                    for key, value in filter.items():
                        if match_filters.get(key):
                            match_filters[key].update(value)
                        else:
                            match_filters[key] = value
                aggregation_statement = [
                    {"$match": match_filters},
                    {"$group": {
                        "_id": None,
                        "min": {"$min": "$" + self.chunking_field},
                        "max": {"$max": "$" + self.chunking_field},
                    }},
                ]
            else:
                aggregation_statement = [{
                    "$group": {
                        "_id": None,
                        "min": {"$min": "$" + self.chunking_field},
                        "max": {"$max": "$" + self.chunking_field},
                    },
                }]
            self.log.info('Getting data range using:\n{0}'.format(
                str(aggregation_statement)
            ))
            minmax = list(collection.aggregate(aggregation_statement))
            if len(minmax) == 0:
                self.log.info('There appears to be no relevant data!')
                self.upload_data([]) # Upload nothing!
                return
            if not len(minmax) == 1:
                raise Exception('Unexpected error occurred')
            minmax = minmax[0]
            current_val = minmax['min'] # == min_val
            max_val = minmax['max']
            self.log.info('loading data from {0} to {1}.'.format(
                str(current_val),
                str(max_val),
            ))
            while current_val < max_val:
                next_val = min(max_val, current_val + self.chunking_interval)
                self.log.info('Getting data from {0} to {1}...'.format(
                    str(current_val),
                    str(next_val),
                ))

                filter_expressions = deepcopy(base_filters)
                lt_or_lte = '$lte' if next_val == max_val else '$lt'
                filter_expressions += [{self.chunking_field:{
                    '$gte': current_val,
                    lt_or_lte: next_val,
                }}]

                filter_expressions = {'$and': filter_expressions}
                self.log.info('Filter expression: {0}'.format(
                    str(filter_expressions)
                ))

                self.log.info('Uploading chunk...')
                self.extract_and_load_paginated(
                    collection=collection,
                    filter_expressions=filter_expressions,
                    page_size=self.pagination_limit,
                    last_id=None,
                )

                current_val = next_val
        else:
            # Just get all data in one go
            if base_filters:
                filter_expressions = {'$and': base_filters}
            else:
                filter_expressions = None
            self.log.info(
                'Filter expression: {0}'.format(str(filter_expressions))
            )
            # Upload data
            self.extract_and_load_paginated(
                collection=collection,
                filter_expressions=filter_expressions,
                page_size=self.pagination_limit,
                last_id=None,
            )



try: # load operators with dependencies installed, ignore others
    from ewah.operators.fx_operator import EWAHFXOperator
except ImportError:
    from ewah.operators.base_operator import EWAHEmptyOperator as EWAHFXOperator
try:
    from ewah.operators.google_analytics_operator import EWAHGAOperator
except ImportError:
    from ewah.operators.base_operator import EWAHEmptyOperator as EWAHGAOperator
try:
    from ewah.operators.mysql_operator import EWAHMySQLOperator
except ImportError:
    from ewah.operators.base_operator import EWAHEmptyOperator as EWAHMySQLOperator
try:
    from ewah.operators.oracle_operator import EWAHOracleSQLOperator
except ImportError:
    from ewah.operators.base_operator import EWAHEmptyOperator as EWAHOracleSQLOperator
try:
    from ewah.operators.postgres_operator import EWAHPostgresOperator
except ImportError:
    from ewah.operators.base_operator import EWAHEmptyOperator as EWAHPostgresOperator
try:
    from ewah.operators.s3_operator import EWAHS3Operator
except ImportError:
    from ewah.operators.base_operator import EWAHEmptyOperator as EWAHS3Operator
try:
    from ewah.operators.facebook_operator import EWAHFBOperator
except ImportError:
    from ewah.operators.base_operator import EWAHEmptyOperator as EWAHFBOperator
try:
    from ewah.operators.google_sheets_operator import EWAHGSpreadOperator
except ImportError:
    from ewah.operators.base_operator import EWAHEmptyOperator as EWAHGSpreadOperator
try:
    from ewah.operators.mongodb_operator import EWAHMongoDBOperator
except ImportError:
    from ewah.operators.base_operator import EWAHEmptyOperator as EWAHGSpreadOperator
try:
    from ewah.operators.shopify_operator import EWAHShopifyOperator
except ImportError:
    from ewah.operators.base_operator import EWAHEmptyOperator as EWAHShopifyOperator
try:
    from ewah.operators.zendesk_operator import EWAHZendeskOperator
except ImportError:
    from ewah.operators.base_operator import EWAHEmptyOperator as EWAHZendeskOperator



from airflow import DAG

from ewah.ewah_utils.airflow_utils import etl_schema_tasks

from datetime import datetime, timedelta
from collections.abc import Iterable
from copy import deepcopy

def dag_factory_drop_and_replace(
        dag_name,
        dwh_engine,
        dwh_conn_id,
        el_operator,
        operator_config,
        target_schema_name,
        target_schema_suffix='_next',
        target_database_name=None,
        default_args=None,
        start_date=datetime(2019, 1, 1),
        schedule_interval=timedelta(days=1),
        end_date=None,
        read_right_users=None,
        additional_dag_args={},
        additional_task_args={},
    ):

    if not hasattr(el_operator, '_IS_FULL_REFRESH'):
        raise Exception('Invalid operator supplied!')
    if not el_operator._IS_FULL_REFRESH:
        raise Exception('Operator does not support full refreshs!')
    if not read_right_users is None:
        if type(read_right_users) == str:
            read_right_users = read_right_users.split(',')
        if not isinstance(read_right_users, Iterable):
            raise Exception('read_right_users must be an iterable or string!')

    dag = DAG(
        dag_name,
        catchup=False,
        default_args=default_args,
        max_active_runs=1,
        schedule_interval=schedule_interval,
        start_date=start_date,
        end_date=end_date,
        **additional_dag_args,
    )

    kickoff, final = etl_schema_tasks(
        dag=dag,
        dwh_engine=dwh_engine,
        dwh_conn_id=dwh_conn_id,
        target_schema_name=target_schema_name,
        target_schema_suffix=target_schema_suffix,
        target_database_name=target_database_name,
        copy_schema=False,
        read_right_users=read_right_users,
        **additional_task_args
    )

    with dag:
        for table in operator_config['tables'].keys():
            table_config = deepcopy(additional_task_args)
            table_config.update(operator_config.get('general_config', {}))
            table_config.update(operator_config['tables'][table] or {})
            table_config.update({
                'task_id': 'extract_load_'+table,
                'dwh_engine': dwh_engine,
                'dwh_conn_id': dwh_conn_id,
                'target_table_name': table,
                'target_schema_name': target_schema_name,
                'target_schema_suffix': target_schema_suffix,
                'target_database_name': target_database_name,
                'drop_and_replace': True,
            })
            table_task = el_operator(**table_config)
            kickoff >> table_task >> final

    return dag



from ewah.operators.base_operator import EWAHBaseOperator
from ewah.ewah_utils.airflow_utils import airflow_datetime_adjustments
from ewah.ewah_utils.python_utils import is_iterable_not_string
from ewah.constants import EWAHConstants as EC

from airflow.hooks.base_hook import BaseHook

from datetime import datetime, timedelta

from requests.auth import HTTPBasicAuth
import requests
import json
import time

class EWAHShopifyOperator(EWAHBaseOperator):

    # template_fields = ()

    _IS_INCREMENTAL = True
    _IS_FULL_REFRESH = False

    _acceptable_api_versions = [
        '2020-07',
    ]
    _current_api_version = '2020-07'
    _base_url = 'https://{shop}.myshopify.com/admin/api/{version}/{object}.json'

    # Set all accepted objects with their potential filter fields with defaults
    #   Note: always expected fields updated_at_max, updated_at_min, limit, id
    #   the key _is_drop_and_replace indicates objects that are only able
    #   to load with a full refresh
    _default_timestamp_fields = ('updated_at_min', 'updated_at_max')
    _accepted_objects = {
        'customers': {
            'ids': None,
            'since_id': None,
        },
        'discount_codes': {
            '_is_drop_and_replace': True,
        },
        'events': {
            '_timestamp_fields': ('created_at_min', 'created_at_max'),
            'since_id': None,
            'filter': None,
            'verb': None,
        },
        'orders': {
            'ids': None,
            'since_id': None,
            'status': 'any',
            'financial_status': None,
            'fulfillment_status': None,
            'fields': None,
        },
        'price_rules': {
            'since_id': None,
        },
        'products': {
            'ids': None,
            'since_id': None,
            'title': None,
            'vendor': None,
            'handle': None,
            'product_type': None,
            'collection_id': None,
            'published_status': None,
        },
    }

    def __init__(self,
        shop_id,
        shopify_object,
        auth_type,
        filter_fields={},
        api_version=None,
        page_limit=250, # API Call pagination limit
    *args, **kwargs):

        if not shopify_object in self._accepted_objects.keys():
            raise Exception('{0} is not in the list of accepted objects!' + \
                ' accepted objects: {1}'.format(
                    shopify_object,
                    ', '.join(self._accepted_objects.keys())
                ))

        if self._accepted_objects[shopify_object].get('_is_drop_and_replace'):
            kwargs['drop_and_replace'] = True

        if not auth_type in ['access_token', 'basic_auth']:
            raise Exception('auth_type must be access_token or basic_auth!')

        if not type(filter_fields) == dict:
            raise Exception('filter_fields must be a dictionary!')
        else:
            for key, value in filter_fields.items():
                if not key in self._accepted_objects['shopify_object'].keys():
                    raise Exception('invalid key {0} in filter fields!'.format(
                        key,
                    ))

        api_version = api_version or self._current_api_version
        if not api_version in self._acceptable_api_versions:
            raise Exception(
                '{0} is not a valid api version! valid versions: {1}'.format(
                    api_version,
                    ', '.join(self._acceptable_api_versions),
                )
            )

        if not type(page_limit) == int or page_limit > 250 or page_limit < 1:
            raise Exception( \
                'Page limit must be a positive integer not exceeding 250!')

        kwargs['primary_key_column_name'] = \
            kwargs.get('primary_key_column_name', 'id')

        # source conn id is not required on operator call level! avoid error
        kwargs['source_conn_id'] = kwargs.get('source_conn_id', '__none__')

        super().__init__(*args, **kwargs)

        self.shop_id = shop_id
        self.shopify_object = shopify_object
        self.auth_type = auth_type
        self.filter_fields = filter_fields
        self.api_version = api_version
        self.page_limit = page_limit

    def execute(self, context):
        # can supply a list of shops - need to run for all shops individually!
        object_metadata = self._accepted_objects[self.shopify_object]
        params = {
            key: val
            for key, val in object_metadata.items()
            if not val is None and not key[:1] == '_'
        }
        params.update(self.filter_fields)
        params.update({'limit': self.page_limit})
        if not self.drop_and_replace:
            timestamp_fields = object_metadata.get(
                '_timestamp_fields',
                self._default_timestamp_fields,
            )
            params.update({
                # Pendulum by coincidence converts to the correct string format
                timestamp_fields[1]: str(context['next_execution_date']),
            })
            if self.test_if_target_table_exists():
                params.update({
                    timestamp_fields[0]: str(context['execution_date']),
                })

        source_conn_id = self.source_conn_id
        auth_type = self.auth_type
        if is_iterable_not_string(self.shop_id):
            # multiple shops to iterate - loop through!
            self.log.info('iterating through multiple shops!')
            for shop_id in self.shop_id:
                # metadata: shop id
                self.log.info('getting data for: {0}'.format(shop_id))
                if hasattr(self.shop_id, 'get') \
                    and hasattr(self.shop_id[shop_id], 'get'):
                    # dict, not list! check for conn details!
                    shop_dict = self.shop_id[shop_id]
                    sci = shop_dict.get('source_conn_id', source_conn_id)
                    at = shop_dict.get('auth_type', auth_type)
                else:
                    sci = source_conn_id
                    at = auth_type
                self._metadata.update({'shop_id': shop_id})
                self.execute_for_shop(context, shop_id, params, sci, at)
        else:
            self._metadata.update({'shop_id': shop_id})
            sci = self.source_conn_id
            at = self.auth_type
            self.execute_for_shop(context, self.shop_id, params, sci, at)

    def execute_for_shop(self,
        context,
        shop_id,
        params,
        source_conn_id,
        auth_type,
    ):
        # Get data from shopify via REST API

        url = self._base_url.format(**{
            'shop': shop_id,
            'version': self.api_version,
            'object': self.shopify_object,
        })

        # get connection for the applicable shop
        conn = BaseHook.get_connection(source_conn_id)
        login = conn.login
        password = conn.password

        if auth_type == 'access_token':
            headers = {
                'X-Shopify-Access-Token': password,
            }
            kwargs_init = {
                'headers': headers,
                'params': params,
            }
            kwargs_links = {'headers': headers}
        elif auth_type == 'basic_auth':
            kwargs_init = {
                'params': params,
                'auth': HTTPBasicAuth(login, password),
            }
            kwargs_links = {'auth': HTTPBasicAuth(login, password)}
        else:
            raise Exception('Authentication type not accepted!')

        # get and upload data
        self.log.info('Requesting data from REST API - url: {0}, params: {1}' \
            .format(url, str(params)))
        r = requests.get(url, **kwargs_init)
        while r.status_code == 200 \
            and r.headers.get('Link') \
            and r.headers['Link'][-10:] == 'rel="next"':
            self.upload_data(json.loads(r.text)[self.shopify_object])
            self.log.info('Requesting next page of data...')
            r = requests.get(r.headers.get('Link')[1:-13], **kwargs_links)

        if not r.status_code == 200:
            raise Exception('Shopify request returned an error {1}: {0}'.format(
                r.text,
                str(r.status_code),
            ))

        # The last page is not uploaded within the while loop!
        self.upload_data(json.loads(r.text)[self.shopify_object])
